// J.DepP -- Japanese Dependency Parsers
//  $Id: pdep.cpp 494 2010-09-22 07:43:31Z ynaga $
// Copyright (c) 2008-2010 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include "pdep.h"

namespace pdep {
  
  void DepParser::_set_lexical_feature (int from, int to, ny::uint &fi) {
    // bit-based assignment of lexical features
    static const ny::uint flag_len = sizeof (flag_t) * 8;
    flag_t * jkbit = new flag_t[_jklen];
    std::fill (&jkbit[0], &jkbit[_jklen], 0);
    // set lexical bit
    for (int k = from; k <= to; ++k) {
      const std::vector <ny::uint> &kv = _s->bun[k].kv;
      // check joshi
      for (std::vector <ny::uint>::const_iterator it = kv.begin ();
           it != kv.end (); ++it)
        jkbit[*it / flag_len] |= (static_cast <flag_t> (1) << (*it % flag_len));
    }
    // set lexical feature
    for (size_t bit_base = 0; bit_base < _jklen; ++bit_base) {
      ny::uint bit_offset = 0;
      while (jkbit[bit_base]) {
        if (jkbit[bit_base] & static_cast <flag_t> (1))
          _fv.push_back (fi + bit_offset);
        jkbit[bit_base] >>= 1;
        ++bit_offset;
      }
      fi += flag_len; // shift bits
    }
    delete [] jkbit;
  }

  void DepParser::_set_global_feature (int i, int j, const Bunsetsu &bi, const Bunsetsu &bj, ny::uint &fi) {
    // activate global feature
    const ny::uint d = j - i;
    int  kakko_in  = 0;
    bool kuten_in  = false;
    bool touten_in = false;
	  
    for (int k = i + 1; k < j; ++k) {
      const Bunsetsu &b = _s->bun[k];
      kakko_in  += b.kakko;
      kuten_in  |= b.kuten;
      touten_in |= b.touten;
    }
    // mod-head: distance / parenthesis / punctuation
    _fv.push_back (d == 1 ? fi : (d > 6 ? fi+1 : fi+2)); fi += 3;
#ifndef USE_EMNLP_FEAT
    _fv.push_back ((kakko_in == 0) ? fi : ((kakko_in > 0) ? fi+1 : fi+2)); fi += 3; // verbose; kv
#endif
    _fv.push_back (! kuten_in     ? fi : fi+1); fi += 2;
    _fv.push_back (! touten_in    ? fi : fi+1); fi += 2;
    // mod: bos  / parenthesis / punctuation
    _fv.push_back (i == 0         ? fi : fi+1); fi += 2;
    _fv.push_back (bi.kakko == 0  ? fi : (bi.kakko > 0 ? fi+1 : fi+2)); fi += 3;
    _fv.push_back (! bi.kuten     ? fi : fi+1); fi += 2;
    _fv.push_back (! bi.touten    ? fi : fi+1); fi += 2;
    // head: eos / parenthesis / punctuation
    _fv.push_back (j == _s->len-1 ? fi : fi+1); fi += 2;
    _fv.push_back (bj.kakko == 0  ? fi : (bj.kakko > 0 ? fi+1 : fi+2)); fi += 3;
    _fv.push_back (! bj.kuten     ? fi : fi+1); fi += 2;
    _fv.push_back (! bj.touten    ? fi : fi+1); fi += 2;
    // particle
    _set_lexical_feature (i+1, j-1, fi);
#ifndef USE_EMNLP_FEAT
    _set_lexical_feature (i, i, fi);
    _set_lexical_feature (j, j, fi);
#endif
  }

  void DepParser::_set_local_feature (const Bunsetsu &bi, const int mid, ny::uint &offset) {
    // set features in a particlular morpheme bi.morph[mid]
    if (mid >= 0) {
      const Morph &m = bi.morph[mid];
      _fv.push_back (MAGIC * m.surf - offset);
      _fv.push_back (MAGIC * m.cpos - offset);
      _fv.push_back (MAGIC * m.fpos - offset);
      _fv.push_back (MAGIC * m.infl - offset);
    }
    ++offset;
  }

  inline bool DepParser::_event_gen_bi (const int i, const int j) {
    _fv.resize (0);
    const Bunsetsu &bi = _s->bun[i]; // modifier candidate
    const Bunsetsu &bj = _s->bun[j]; // head     candidate
    
    ny::uint fi = _global_offset;
    _set_global_feature (i, j, bi, bj, fi);
  
    ny::uint offset = 0;
    _set_local_feature (bi, bi.head, offset);
    _set_local_feature (bi, bi.tail, offset);
#ifndef USE_EMNLP_FEAT
    _set_local_feature (bj, 0,       offset);
#endif
    _set_local_feature (bj, bj.head, offset);
    _set_local_feature (bj, bj.tail, offset);
#ifndef USE_EMNLP_FEAT
    // A07
    if (j != _s->len-1) { // head+1
      // 0 +0.1
      const Bunsetsu &bk = _s->bun[j+1];
      _fv.push_back (MAGIC * bk.morph[0].surf - offset);
      ++offset;
      if (bk.head != -1) _fv.push_back (MAGIC * bk.morph[bk.head].surf - offset);
      ++offset; 
      if (bk.tail != -1) _fv.push_back (MAGIC * bk.morph[bk.tail].surf - offset);
      ++offset;
      _fv.push_back (MAGIC * bk.morph[bk.mlen-1].surf - offset);
      ++offset;
    } else {
      offset += 4;
    }
#endif
    if (_mode != parse) std::sort (_fv.begin (), _fv.end ()); // for ME
    return true;
  }
  
  inline bool DepParser::_event_gen_tri (const int i, const int j, const int k) {
    _fv.resize (0);
  
    const Bunsetsu &bi = _s->bun[i]; // modifier cand
    const Bunsetsu &bj = _s->bun[j]; // head cand1
    const Bunsetsu &bk = _s->bun[k]; // head cand2

    ny::uint fi = _global_offset;
    _set_global_feature (i, j, bi, bj, fi);
    _set_global_feature (i, k, bi, bk, fi);
    ny::uint offset = 0;
    _set_local_feature (bi, bi.tail,   offset);
    _set_local_feature (bi, bi.head,   offset);
    _set_local_feature (bj, 0,         offset);
    _set_local_feature (bj, bj.head,   offset);
    _set_local_feature (bj, bj.tail,   offset);
    _set_local_feature (bk, 0,         offset);
    _set_local_feature (bk, bk.head,   offset);
    _set_local_feature (bk, bk.tail,   offset);
    // A07
    if (j != _s->len-1) {
      const Bunsetsu &bl = _s->bun[j+1];
      _fv.push_back (MAGIC * bl.morph[0].surf - offset);
      ++offset;
      if (bl.head != -1) _fv.push_back (MAGIC * bl.morph[bl.head].surf - offset);
      ++offset; 
      if (bl.tail != -1) _fv.push_back (MAGIC * bl.morph[bl.tail].surf - offset);
      ++offset;
      _fv.push_back (MAGIC * bl.morph[bl.mlen-1].surf - offset);
      ++offset;
    } else {
      offset += 4;
    }
    if (_mode != parse) std::sort (_fv.begin (), _fv.end ()); // for ME
    return true;
  }

  inline void DepParser::_processSample (bool flag) {
    if (_learner == OLL) {
#ifdef USE_PA
      opal::ex_t x;
      _pa->set_ex (x, flag ? +1 : -1, _fv, true, _d > 0);
      _ex_pool.put (x);
#endif
    }
    else if (_learner == ME) {
#ifdef USE_ME
      ME_Sample ms (flag ? "+1" : "-1");
      _setExpanded (_fv, _d, ms);
      _libme.add_training_sample (ms);
#endif
    }
  }

  bool DepParser::_eventOutputTournament () {
    const int len = _s->len;
    for (int i = 0; i < len-2; ++i) {
      int h = _s->bun[i].did; // head
      for (int j = i+1; j <= len-1; ++j) {
        bool flag = true;
        if      (j < h) { flag = true;  _event_gen_tri (i, j, h); }
        else if (j > h) { flag = false; _event_gen_tri (i, h, j); }
        else continue;
        ++_cnum;
        _processSample (flag);
        _writeFv (flag);
      }
    }
    return true;
  }

  template <process_t MODE>
  bool DepParser::_processLinear () {
    const int len = _s->len;
    for (int j = 1; j < len; ++j) {
      _stack.push (j - 1);
      while (! _stack.empty ()) {
        const int i = _stack.top ();
        if (j != len - 1) {
          _event_gen_bi (i, j);
          ++_cnum;
          bool flag = (j == _s->bun[i].did);
          // output example for training / fstrie construction
          if (MODE != parse) _writeFv (flag);
          if (MODE == learn) _processSample (flag);     // learn
          else               flag = _binClassify (_fv); // parse
          if (! flag) break;
        }
        _s->bun[i].est_did = j;
        _stack.pop ();
      }
    }
    return true;
  }
  

  template <process_t MODE>
  bool DepParser::_processChunking () {
    const int len = _s->len;
    for (int i = 0; i < len - 1; ++i)
      _cinfo.push_back (chunk_info (i));
    // _cinfo.push_back (chunk_info (i, false)); // deterministic
    while (! _cinfo.empty ()) {
      std::list <chunk_info>::reverse_iterator it = _cinfo.rbegin ();
      bool next = true; // D
      int i = -1;
      int j = it->id;
      _s->bun[j].est_did = len - 1;
      for (++it; it != _cinfo.rend (); ++it) {
        i = it->id;
        // if (! it->done) { // deterministic (assuming no dynamic features)
          _event_gen_bi (i, j);
          bool flag = (j == _s->bun[i].did);
          if (MODE != parse) _writeFv (flag);
          if (MODE == learn) _processSample (flag);
          else               flag = _binClassify (_fv);
          if (flag) _s->bun[i].est_did = j;
          // it->done = true; // deterministic
        // }
        if (_s->bun[i].est_did == -1 && next) { // O & D
          it = std::list <chunk_info>::reverse_iterator (_cinfo.erase (it.base ()));
          // it->done = false; // deterministic
        }
        next = _s->bun[i].est_did != -1;
        j = i;
      }
      if (_s->bun[i].est_did != -1) _cinfo.erase (it.base ());
    }
    return true;
  }

  template <process_t MODE>
  bool DepParser::_processBackward () {
    const int len = _s->len;
    for (int i = len - 2; i >= 0; --i) {
      ny::fl_t max_prob = 0;
      for (int j = i + 1; j != len - 1; j = _s->bun[j].est_did) { // deterministic
        ny::fl_t prob = 0;
        _event_gen_bi (i, j);
        ++_cnum;
        bool flag = (j == _s->bun[i].did);
        // output example for training / fstrie construction
        if (MODE != parse) _writeFv (flag);
        if (MODE == learn) { // learn
          _processSample (flag);
          if (flag) { _s->bun[i].est_did = j; prob = 1.0; }
            // { _s->bun[i].est_did = j; prob = 1.0; break; } // deterministic
        } else {
          prob = _classify (_fv);
        }
        // if (prob >= 0.5) { _s->bun[i].est_did = j; break; } // deterministic
        if (prob > max_prob) { _s->bun[i].est_did = j; max_prob = prob; }
      }
    }
    return true;
  }
  
  bool DepParser::_parseTournament () {
    int len = _s->len;
    for (int i = len-2; i >= 0; --i) {
      int did = i+1; // head
      int j = _s->bun[did].est_did; // head of head
      while (j != -1) {
        _event_gen_tri (i, did, j);
        ++_cnum;
        if (_mode == cache) _writeFv (did < _s->bun[i].did);
        bool flag = _binClassify (_fv);
        if (flag) did = j; // RIGHT
        j = _s->bun[j].est_did;
      }
      _s->bun[i].est_did = did;
    }
    return true;
  }
  
  template <process_t MODE>
  inline bool DepParser::_process () {
    switch (_parser) {
      case linear:
        switch (MODE) {
          case parse: _processLinear <parse> (); break;
          case cache: _processLinear <cache> (); break;
          case learn: _processLinear <learn> (); break;
          case dict: break;
        }
        break;
      case chunking:
        switch (MODE) {
          case parse: _processChunking <parse> (); break;
          case cache: _processChunking <cache> (); break;
          case learn: _processChunking <learn> (); break;
          case dict: break;
        }
        break;
      case backward:
        switch (MODE) {
          case parse: _processBackward <parse> (); break;
          case cache: _processBackward <cache> (); break;
          case learn: _processBackward <learn> (); break;
          case dict: break;
        }
        break;
      case tournament: // asymmetry
        if (MODE == learn) _eventOutputTournament (); else _parseTournament ();
    }
    if (_s->len >= 1) { // ignore sentene w/ a single bunsetsu
      ++_snum;
      _bnum += _s->len-1;
      if (MODE != dict && MODE != learn) { // calculate statistics
        const int len = _s->len;
        char _flag = 0x3; // 11
        for (int i = 0; i < len - 1; ++i) {
          const Bunsetsu &b = _s->bun[i];
          if (b.est_did == b.did) ++_bcorr; else _flag &= 0x1;   // 01
          // std::cout << (b.est_did == b.did ? 1 : 0) << std::endl; // McNemar
          if (_verbosity > 1)
            { if (b.did == i + 1) ++_bbase; else _flag &= 0x2; } // 10
        }
        if (_flag & 0x2) ++_scorr;
        // std::cout << (_flag & 0x2 ? 1 : 0) << std::endl; // McNemar
        if (_verbosity > 1 && _flag & 0x1) ++_sbase;
      }
    }
    return true;
  }
  
  template <process_t MODE>
  bool DepParser::_analyze () {
    std::string afn; // the target file name to analyze
    std::string mode;
    switch (MODE) {
      case dict:  mode = "dict";  afn += _opt["-lf"]; break;
      case learn: mode = "learn"; afn += _opt["-lf"]; break;
      case cache: mode = "cache"; afn += _opt["-cf"]; break;
      case parse: mode = "parse"; afn += _opt["-tf"]; break;
      default: break;
    }
    const bool output =
      (_mode == parse && (_opt["-tf"] == "-" || _verbosity == -1));
    if (MODE != dict) setSentence (output);

    FILE* reader = (afn == "-") ? stdin : fopen (afn.c_str (), "r");
    if (! reader) ny::print_err (HERE "no such file: %s\n", afn.c_str ());
    _bnum = _snum = _cnum = 0;
    ny::uint n  = 0;
    char*  line = 0;
    size_t read = 0;
    while (ny::getLine (reader, line, read)) { // bunsetsu chunking
      switch (*line) {
        case '\0': break;
        case '*': if (MODE != dict) _s->addBunsetsu (line); break; // bunsetsu start
        case '#': // sentence start
          if (MODE != dict) {
            if (read >= 4 && line[2] == 'S') {
#ifdef USE_TIMER
              if (MODE != learn) _preproc_t->startTimer ();
#endif
              _s->clear ();
              if (output) _s->setHeader (line, read - 1);
            } else {
              write (1, line, read - 1);
              write (1, "\n", 1);
            }
          }
          break;
        case 'E': // end of sentence
          if (MODE != dict) {
            _s->setInfo (); // bug?
#ifdef USE_TIMER
            if (MODE != learn) {
              _preproc_t->stopTimer ();
              _parser_t->startTimer ();
            }
#endif
            _process <MODE> ();
            if (output) {
#ifdef USE_TIMER
              _io_t->startTimer ();
              _s->printParsed ();
              _io_t->stopTimer ();
#else
              _s->printParsed ();
#endif
            }
#ifdef USE_TIMER
            if (MODE != learn) _parser_t->stopTimer ();
#endif
          }
          if (MODE != parse) {
            if (++n % 1000 == 0)
              std::cerr << "\r" << mode << ": " << n << " sent. processed";
            if ((MODE == cache && _lim_cache && n >= _lim_cache) ||
                (MODE == learn && _lim && n >= _lim) ||
                (MODE == dict  && _lim && n >= _lim)) {
              if (_s) { delete _s; _s = NULL; }
              if (reader!= stdin) std::fclose (reader);
              return true;
            }
          }
          break;
        default: // morpheme
          if (MODE == dict) _registerMorph (line, read - 1);
          else              _s->bun[_s->len - 1].addMorph (line, read - 1);
      }
    }
    if (_s) { delete _s; _s = NULL; }
    if (reader != stdin) std::fclose (reader);
    if (MODE != parse)
      std::cerr << "\r" << mode << ": " << n << " sent. processed; done." << std::endl;
    return true;
  }

  bool DepParser::_learn () {
    switch (_learner) {
#ifdef USE_PA
      case OLL:
        _pa->train (_ex_pool, _paiter);
        _pa->save (_mfn.c_str ());
        delete _pa;
        break;
#endif
#ifdef USE_ME
      case ME:
        _libme.train (_cthr, _me_t == L2 ? _rcost : 0.0, _me_t == L1 ? _rcost : 0.0);
        _libme.save_to_file (_mfn);
        break;
#endif
#ifdef USE_SVM
      case SVM: {
        TinySVM::Example *ex = new TinySVM::Example;
        ex->read (_lfn.c_str ());
        std::ostringstream ss;
        ss << "-c " << _opt["-rc"] << " -t 1 -d " << _opt["-d"]
           << " -s " << _opt["-s"] << " -r " << _opt["-r"];
        _tinysvm = ex->learn (ss.str ().c_str ());
        _tinysvm->write (_mfn.c_str ());
        delete ex;
        delete _tinysvm;
        break;
      }
#endif
      default: assert (! "learner disabled");
    }
    return true;
  }

  bool DepParser::_init () {
#ifdef USE_TIMER
    if (_verbosity > 0)
      fprintf (stderr, "clock=%.6Lfms.\n", ny::Timer::clock);
    _dict_t->startTimer ();
    _setMorphDic ();
    _dict_t->stopTimer ();
#else
    _setMorphDic ();
#endif
    // magic parameters
    _global_offset = MAGIC * _word_id + 1;
    _jklen = static_cast <ny::uint> ((_jk2id.size () + 1) /
                                     (sizeof (flag_t) * 8)) + 1;
    return true;
  }
  
  bool DepParser::run () {
    if (_mode != parse && _mode != cache) { // learn or both
#ifdef USE_PA
      if (_learner == OLL)
        _pa = new opal::Model (_d, _d ? PA_C : 0, PA_MODEL, PA_AVE, PA_S, static_cast <opal::fl_t> (_rcost), true); // see pa.h
#endif
      _writer = std::fopen (_lfn.c_str (), "wb");
      char buf[1024*2048];
      std::setvbuf (_writer, buf, _IOFBF, 1024 * 2048);
      _analyze <learn> ();
      std::fclose (_writer);
      _learn ();
    }
    if (_mode != learn) { // parse, cache or both
      if (_mode == both) _mode = parse;
      FILE* reader = std::fopen (_mfn.c_str (), "rb");
      if (! reader) ny::print_err (HERE "no model file: %s\n", _mfn.c_str ());
      std::fclose (reader);
      // setup classifier
      parseInit ();
      if (_mode == parse)
        _analyze <parse> ();
      else {
        _writer = std::fopen (_cfn.c_str (), "wb");
        char buf[1024*2048];
        std::setvbuf (_writer, buf, _IOFBF, 1024 * 2048);
        _analyze <cache> ();
        std::fclose (_writer);
      }
      // print statistics
      _printStatistics ("ACCURACY", _bcorr, _bnum, _scorr, _snum);
      if (_verbosity > 1) _printStatistics ("BASELINE", _bbase, _bnum, _sbase, _snum);
      // print timings
#ifdef USE_TIMER
      _timer_pool.print ();
#endif
      if (_classifier == pME) {
#ifdef USE_ME
#ifdef USE_PROFILE
        _me->printHitRatio ();
#endif
#ifdef USE_TIMER
        _me->printTimer ();
#endif
        delete _me;
#endif
      } else if (_classifier == pSVM) {
#ifdef USE_SVM
#ifdef USE_PROFILE
        _svm->printHitRatio ();
#endif
#ifdef USE_TIMER
        _svm->printTimer ();
#endif
        delete _svm;
#endif
      } else if (_classifier == PA) {
#ifdef USE_PA
        delete _pa;
#endif
      }
#ifndef NDEBUG
#ifdef USE_SVM
      if (_classifier == TinySVM) delete _tinysvm;
#endif
#endif
    }
    return true;
  }
  // register each morphemes to dict
  bool DepParser::_registerMorph (char* cs, const ssize_t &len) {
    ny::uint i = 0;
    char* surf = 0;
    const char* const pend = cs + len;
    bool free_flag = true;
    for (char* p = cs; cs < pend && i < NUM_FIELD; ++i) {
      p = ny::skipChar (p, ' ', pend);
      switch (i) {
        case YOMI: break;
        case TYPE: break;
        case FIN:  break;
        case INFL: break;
        case SURF:
          surf = new char[p - cs]; // later delete from _jk_set
          std::strcpy (surf, cs); break;
        case CPOS:
          if (std::strcmp (cs, POST_PARTICLE) == 0)
            if (_jk_set.insert (surf).second) free_flag = false;
          break;
#ifdef USE_EMNLP_FEAT
        case FPOS:
          if (std::strcmp (cs, JUMAN_BRACKET_START) == 0 ||
              std::strcmp (cs, JUMAN_BRACKET_END)   == 0)
            if (_jk_set.insert (surf).second) free_flag = false;
          break;
#endif
      }
      char* s = new char[p - cs]; // delete from w2wi
      std::strcpy (s, cs);
      std::pair <dic_t::iterator, bool> itb =
        _w2wi[i].insert (dic_t::value_type (s, word ()));
      if (! itb.second) {
        delete [] s;
        ++itb.first->second.count;
      }
      cs = p;
    }
    if (i < NUM_FIELD) ny::print_err (HERE "illegal field.\n");
    if (free_flag) delete [] surf;
    return true;
  }
  // morphological dictionary are extracted from the training data
  bool DepParser::_setMorphDic () {
    FILE* reader = std::fopen (_dfn.c_str (), "r");
    if (_verbosity > 0) std::cerr << "Loading dict..";
    // first check whether the .dic has been already built.
    if (! reader) {
      if (_verbosity > 0) std::cerr << "not found: " << _dfn << std::endl;
      _analyze <dict> ();
      // assign id from coarse to fine field
      ny::counter <ny::uint>::type dic_counter;
      for (ny::uint i = 0; i < NUM_FIELD; ++i)
        dic_counter.push_back (std::make_pair (_w2wi[i].size (), i));
      std::sort (dic_counter.begin (), dic_counter.end ());
      ny::uint total = 0;
      for (ny::uint i = 0; i < NUM_FIELD; ++i) {
        const ny::uint field_id = dic_counter[i].second;
        dic_t &dic = _w2wi[field_id];
        if (_verbosity > 0) std::cerr << " propcssing field " << field_id << "..";
        ny::counter <const char*>::type word_counter;
        word_counter.reserve (dic.size ());
        for (dic_t::const_iterator dit = dic.begin (); dit != dic.end (); ++dit)
          word_counter.push_back (ny::counter <const char*>::type::value_type
                                  (dit->second.count, dit->first));
        std::sort (word_counter.rbegin (), word_counter.rend (),
                   pointer_counter_less <char> ());
        ny::uint deleted = 0;
        ny::uint j = 0;
        for (ny::counter <const char*>::type::const_iterator wit = word_counter.begin ();
             wit != word_counter.end (); ++wit) {
          if (++j % 100 == 0)
            if (_verbosity > 0)
              std::cerr << "\r processing field " << field_id << ".."
                        << j << "/" << dic.size () << " completed";
          const char* str = wit->second;
          word &w = dic[str];
          w.id = _word_id;
          if (w.count >= _cthr) {
            ny::uint c[1]; *c = _word_id;
            _dakey_ary[field_id]->push_back (new DictKey (const_cast <char*> (str), c)); // delete from DA
            ++total;
          } else {
            ++deleted;
          }
          ++_word_id;
        }
        { // insert key for rare/unknown
          word rare (_word_id, 0);
          dic.insert (dic_t::value_type (UNKNOWN, rare));
          ny::uint c[1]; *c = _word_id;
          _dakey_ary[field_id]->push_back (new DictKey (const_cast <char*> (UNKNOWN), c)); // delete from DA
          ++_word_id;
        }
        if (field_id == SURF) { // jk etc.
          FILE* writer = std::fopen (_jfn.c_str (), "wb");
          ny::counter <const char*>::type id2jo;
          // cut morpheme with frequency < _cthr
          for (charp_set::iterator it = _jk_set.begin (); it != _jk_set.end (); ++it)
            id2jo.push_back (ny::counter <const char*>::type::value_type
                             (dic[*it].count, *it));
          std::sort (id2jo.rbegin (), id2jo.rend (),
                     pointer_counter_less <char> ()); // frequent to rare
          for (ny::uint l = 0; l < id2jo.size (); ++l) {
            const char     *s = id2jo[l].second;
            const ny::uint c  = dic[s].count;
            const ny::uint jo = dic[s].id;
            if (c >= _cthr) _jk2id.insert (Jmap::value_type (jo, l));
            std::fwrite (&jo, sizeof (ny::uint), 1, writer);
            std::fwrite (&l,  sizeof (ny::uint), 1, writer);
            std::fwrite (&c,  sizeof (ny::uint), 1, writer);
            delete [] s;
          }
          std::fclose (writer);
        }
        if (_verbosity > 0)
          std::cerr << "\r propcssing field " << field_id << ".." << j << " => "
                    << dic.size () - deleted << " completed" << std::endl;
      }
      --_word_id;
      if (_verbosity > 0)
        std::cerr << total << " strings (freq >= " << _cthr << ") found " << std::endl;

      FILE* writer = std::fopen (_dfn.c_str (), "w");
      for (ny::uint i = 0; i < NUM_FIELD; ++i) {
        dic_t &dic = _w2wi[i];
        for (dic_t::const_iterator dit = dic.begin (); dit != dic.end (); ++dit) {
          const char* key  = dit->first;
          const word &word = dit->second;
          std::fprintf (writer, "%d %s %d %d\n", i, key, word.id, word.count);
        }
      }
      std::fclose (writer);
    } else {
      // only when dictionary has been already built
      if (_verbosity > 0) std::cerr << "cache found!" << std::endl;
      if (_verbosity > 0) std::cerr << "Loading dict DA..";
      
      FILE* joshi_reader = std::fopen (_jfn.c_str (), "rb");
      ny::uint jo = 0;
      // eliminate things as well
      while (std::fread (&jo, sizeof (ny::uint), 1, joshi_reader)) {
        ny::uint j = 0;
        ny::uint c = 0;
        std::fread (&j, sizeof (ny::uint), 1, joshi_reader);
        std::fread (&c, sizeof (ny::uint), 1, joshi_reader);
        if (c >= _cthr) _jk2id.insert (Jmap::value_type (jo, j));
      }
      std::fclose (joshi_reader);
      
      bool flag = true;
      std::ostringstream ss;
      ss << _mbase << ".e" << _cthr << ".d" << ny::DA_SUFFIX;
      std::string dafn = ss.str ();
      FILE * dda_reader = std::fopen (dafn.c_str (), "rb");
      if (! dda_reader) {
        flag = false;
      } else {
        if (std::fseek (dda_reader, 0, SEEK_END) != 0) return -1;
        ny::uint da_size_total = std::ftell (dda_reader) - sizeof (ny::uint) * 2 * NUM_FIELD;
        if (std::fseek (dda_reader, 0, SEEK_SET) != 0) return -1;
        _ddap = new char[da_size_total];
        ny::uint da_offset = 0;
        for (ny::uint i = 0; i < NUM_FIELD; ++i) {
          ny::uint size    = 0;
          ny::uint da_size = 0;
          if (std::fread (&size,    sizeof (ny::uint), 1, dda_reader) &&
              std::fread (&da_size, sizeof (ny::uint), 1, dda_reader) &&
              std::fread (_ddap + da_offset, sizeof (char), da_size, dda_reader)) {
            _dict_da[i]->set_array (_ddap + da_offset, da_size / _dict_da[i]->unit_size ());
            da_offset += da_size;
            _word_id += static_cast <ny::uint> (size);
          } else {
            flag = false;
            break;
          }
        }
      }
      std::fclose (dda_reader);
      if (flag) {
        --_word_id;
        if (_verbosity > 0)
          std::cerr << "done; max id=" << _word_id << std::endl;
        return true;
      } else { // build dict DA
        if (_verbosity > 0) std::cerr << "broken!" << std::endl;
        ny::uint total = 0;
        char*    line  = 0;
        size_t   read  = 0;
        while (ny::getLine (reader, line, read)) {
          ny::uint i = ny::strton <ny::uint> (std::strtok (line, " ")); // dic id
          char* cs = std::strtok (NULL, " ");
          char* s = new char[std::strlen (cs) + 1]; // delete from w2wi
          std::strcpy (s, cs);
          ny::uint fi  = ny::strton <ny::uint> (std::strtok (NULL, " "));
          ny::uint count = ny::strton <ny::uint> (std::strtok (NULL, " "));
          if (fi > _word_id) _word_id = fi;
          if (std::strcmp (s, UNKNOWN) == 0 || count >= _cthr) {
            ++total;
            ny::uint c[1]; *c = fi; // delete from da
            _dakey_ary[i]->push_back (new DictKey (cs, c)); // delete from DA
          }
          _w2wi[i].insert (dic_t::value_type (s, word (fi, count)));
        }
        if (_verbosity > 0)
          std::cerr << "done; " << total << " strings (freq >= " << _cthr << ") "
                    << "found " << std::endl;
      }
    }
    if (reader) std::fclose (reader);
    if (_verbosity > 0) std::cerr << "Building dict DA.." << std::endl;
    _word_id = FI_OFFSET; // bug!! set appropriate  _word_id for e > 0

    std::ostringstream ss;
    ss << _mbase << ".e" << _cthr << ".d" << ny::DA_SUFFIX;
    std::string dafn = ss.str ();
    ss.str ("");
    for (ny::uint i = 0; i < NUM_FIELD; ++i) {
    
      std::vector <DictKey*>* dakey = _dakey_ary[i];
      DictKeypLess dictkeypless;
      std::sort (dakey->begin (), dakey->end (), dictkeypless);
    
      std::vector <const char*> str;
      std::vector <size_t> len;
      std::vector <int> val;
      for (std::vector <DictKey*>::const_iterator dit = dakey->begin ();
           dit != dakey->end (); ++dit) {
        str.push_back ((*dit)->key);
        len.push_back ((*dit)->len);
        val.push_back (*((*dit)->cont));
      }
      FILE* writer = std::fopen (dafn.c_str (), "ab");
      ny::uint size = static_cast <ny::uint> (dakey->size ());
      _word_id += static_cast <ny::uint> (size); // bug!!
      std::fwrite (&size, sizeof (ny::uint), 1, writer);
      ss << "dict (field " << i << ")";
      std::string name (ss.str ());
      if (_verbosity > 0) std::cerr << " Building " << name << " DA..";
      ss.str ("");
      ny::DA da;
      if (da.build (str.size (), &str[0], &len[0], &val[0]) != 0)
        ny::print_err (HERE "failed to build DA.\n");
      ny::uint da_size = da.size () * da.unit_size ();
      std::fwrite (&da_size, sizeof (ny::uint), 1, writer);
      std::fclose (writer);
      if (da.save (dafn.c_str (), "ab") != 0)
        ny::print_err (HERE "failed to save DA.\n");
      if (_verbosity > 0) std::cerr << "done." << std::endl;
    }
    reader = std::fopen (dafn.c_str (), "rb");
    if (std::fseek (reader, 0, SEEK_END) != 0) return -1;
    ny::uint da_size_total = std::ftell (reader) - sizeof (ny::uint) * 2 * NUM_FIELD;
    if (std::fseek (reader, 0, SEEK_SET) != 0) return -1;
    _ddap = new char[da_size_total];
    ny::uint size      = 0;
    ny::uint da_size   = 0;
    ny::uint da_offset = 0;
    for (ny::uint i = 0; i < NUM_FIELD; ++i)
      if (std::fread (&size,    sizeof (ny::uint), 1, reader) &&
          std::fread (&da_size, sizeof (ny::uint), 1, reader)) {
        std::fread (_ddap + da_offset, sizeof (char), da_size, reader);
        _dict_da[i]->set_array (_ddap + da_offset, da_size / _dict_da[i]->unit_size ());
        da_offset += da_size;
      }
    --_word_id;
    if (_verbosity > 0)
      std::cerr << "done; max string id=" << _word_id << std::endl;
    
    return true;
  }
  void DepParser::process () { _process <parse> (); }
}
// introduce dynamic progamming into event gen to calculate gap feature
