// J.DepP -- Japanese Dependency Parsers
//  $Id: pdep.h 494 2010-09-22 07:43:31Z ynaga $
// Copyright (c) 2008-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef PDEP_H
#define PDEP_H

#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <dirent.h>
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cerrno>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <list>
#include <stack>
#include <map>
#include <set>
#include "typedef.h"
#include "jdepp_conf.h"
#ifdef USE_HASH
#include <tr1/unordered_map>
#endif
#ifdef USE_DARTS
#include <darts.h>
#endif
#ifdef USE_DARTS_CLONE
#include <darts-clone.h>
#endif
#ifdef USE_DDA
#include "dda.h"
#endif
#include "opt_parse.h"
#ifdef USE_TIMER
#include "timer.h"
#endif
#ifdef USE_SVM
#include <tinysvm.h>
#include "polyk.h"
#endif
#ifdef USE_ME
#include "maxent.h"
#include "llm.h"
#endif
#ifdef USE_PA
#define USE_PRUNE
#define USE_PMT
#include "opal_conf.h"
#include "pa.h"
#include "polyk.h"
#endif

namespace pdep {
  // type alias
  typedef uint64_t flag_t;
  enum field_t    {SURF, YOMI, FIN, CPOS, FPOS, TYPE, INFL, NUM_FIELD};
  enum process_t  {dict, learn, cache, parse, both}; // both: learn and parse
  // static const
  static const char* UNKNOWN = "_";
#ifdef USE_EMNLP_FEAT
  static const ny::uint  MAGIC     = 16; // 8
#else
  static const ny::uint  MAGIC     = 32; // 16
#endif
#ifdef USE_PA
  static const ny::uint     PA_C     = 1000;      // C   = 1000
  static const opal::algo_t PA_MODEL = opal::PA1; // algo = PA1
  static const bool         PA_AVE   = false;     // ave = 1
  static const bool         PA_S     = false;     // r   = 1
#endif
  static const ny::uint  FI_OFFSET = 1;
  // dictionary utility variables
  struct word {
  public:
    ny::uint count;
    int id;
    word () : count (1), id (-1) {}
    word (ny::uint c, int i) : count (c), id (i) {}
  };
  typedef std::map <const char*, word, ny::pless <char> > dic_t;
  typedef ny::map <ny::uint, ny::uint>::type              Jmap;
  // char* -> int; word -> ID (int)
  typedef std::set <char*, ny::pless <char> > charp_set;
  typedef ny::DAKeyBase  <char, ny::uint>  DictKey; // token -> id DA
  typedef ny::DAKeypLess <char, ny::uint>  DictKeypLess;
  // 
  template <typename T>
  struct pointer_counter_less {
    bool operator() (const std::pair <ny::uint, const T*> &a,
                     const std::pair <ny::uint, const T*> &b) const {
      if      (a.first < b.first) return true;
      else if (a.first > b.first) return false;
      return ny::pless <T> () (a.second, b.second);
    }
  };
  
  class DicInfo {
  public:
    ny::DA**       dict_da;
    const ny::uint touten_id;
    const ny::uint kuten_id;
    const ny::uint jo_id;
    const ny::uint kakkos_id;
    const ny::uint kakkoe_id;
#ifdef USE_EMNLP_FEAT
    const ny::uint suf_id;
#endif
    const ny::uint sp_id;
    // rare ids
    ny::uint rare[NUM_FIELD];
  
    DicInfo (ny::DA** _dict_da) :
      dict_da (_dict_da),
      touten_id (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (COMMA)),
      kuten_id  (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (PERIOD)),
      jo_id  (dict_da[CPOS]->exactMatchSearch <ny::DA::result_type> (POST_PARTICLE)),
#if defined (USE_JUMAN_POS)
      kakkos_id (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (JUMAN_BRACKET_START)),
      kakkoe_id (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (JUMAN_BRACKET_END)),
#ifdef USE_EMNLP_FEAT
      suf_id (dict_da[CPOS]->exactMatchSearch <ny::DA::result_type> (JUMAN_SUFFIX)),
#endif
      sp_id  (dict_da[CPOS]->exactMatchSearch <ny::DA::result_type> (JUMAN_SPECIAL))
#elif defined (USE_IPA_POS)
      kakkos_id (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (IPA_BRACKET_START)),
      kakkoe_id (dict_da[FPOS]->exactMatchSearch <ny::DA::result_type> (IPA_BRACKET_END)),
      sp_id  (dict_da[CPOS]->exactMatchSearch <ny::DA::result_type> (IPA_SPECIAL))
#endif
    {
      for (ny::uint i = 0; i < NUM_FIELD; ++i)
        rare[i] =  dict_da[i]->exactMatchSearch <ny::DA::result_type> (UNKNOWN);
    }
  };

  class Morph {
  private:
    const DicInfo* _dinfo;
    ny::DA**       _dict_da;
    int            _verbosity;
    bool           _output;
  public:
    ny::uint surf;
    ny::uint cpos;
    ny::uint fpos;
    ny::uint infl;
    // ny::uint fin;
    // ny::uint yomi;
    // for output
    char   surf_str[64];
    char   fin_str[64];
    char   info_str[1024];
    size_t info_len;
    
    Morph (const DicInfo* dinfo, ny::uint verbosity, bool output): _dinfo (dinfo), _dict_da (dinfo->dict_da), _verbosity (verbosity), _output (output), surf (0), cpos (0), fpos (0), infl (0), info_len (0)
    { surf_str[0] = '\0'; fin_str[0] = '\0'; info_str[0] = '\0'; };
    Morph (const Morph& m) : _dinfo (m._dinfo), _dict_da (m._dict_da), _verbosity (m._verbosity), _output (m._output), surf (m.surf), cpos (m.cpos), fpos (m.fpos), infl (m.infl), info_len (m.info_len) {
      // copy pointers
      if (_output) std::memcpy (info_str, m.info_str, info_len);
      if (_verbosity > 1) {
        std::strcpy (surf_str, m.surf_str);
        std::strcpy (fin_str, m.fin_str);
      }
    };
    ~Morph () {}
    void setInfo (char* cs, const ssize_t &len) {
      if (_output) { info_len = len; std::memcpy (info_str, cs, info_len); }

      ny::uint i = 0;
      const char* const pend = cs + len;
      for (char* p = cs; cs < pend && i < NUM_FIELD; ++i) {
        p = ny::skipChar (p, ' ', pend);
        switch (i) {
          case YOMI: break;
          case TYPE: break;
          case FIN:  if (_verbosity > 1) std::strcpy (fin_str,  cs); break;
          case SURF: if (_verbosity > 1) std::strcpy (surf_str, cs);
          default:
            int n = _dict_da[i]->exactMatchSearch <ny::DA::result_type> (cs, p - cs - 1);
            switch (i) {
              case SURF: surf = (n >= 0) ? n : _dinfo->rare[i]; break;
              case CPOS: cpos = (n >= 0) ? n : _dinfo->rare[i]; break;
              case FPOS: fpos = (n >= 0) ? n : _dinfo->rare[i]; break;
              case INFL: infl = (n >= 0) ? n : _dinfo->rare[i]; break;
            }
        }
        cs = p;
      }
      if (i < NUM_FIELD) ny::print_err (HERE "illegal field.\n");
    }
    bool joshi () { return cpos == _dinfo->jo_id; }
  };

  class Bunsetsu {
  private:
    const DicInfo* _dinfo;
    const Jmap*    _jk2id;
    int            _verbosity;
    bool           _output;
  public:
    std::vector <Morph>    morph;
    std::vector <ny::uint> kv;
    int  head;
    int  tail;
    int  id;
    int  did;
    int  kakko;
    int  est_did;
    bool touten;
    bool kuten;
    int  mlen;
    int  capacity;
    // char dtype;
    Bunsetsu (const DicInfo* dinfo, const Jmap* jk2id, ny::uint verbosity, bool output) : _dinfo (dinfo), _jk2id (jk2id), _verbosity (verbosity), _output (output), morph (), kv (), head (-1), tail (-1), id (-1), did (-1), kakko (0), est_did (-1), touten (false), kuten (false), mlen (0), capacity (1)
    { morph.resize (capacity, Morph (dinfo, _verbosity, _output)); }

    void clear () { // leave other members overwritten
      kv.resize (0);
      head   = tail  = est_did = -1;
      kakko  = mlen  = 0;
      touten = kuten = false;
    }
    Bunsetsu (const Bunsetsu &b) : _dinfo (b._dinfo), _jk2id (b._jk2id), _verbosity (b._verbosity), _output (b._output), morph (b.morph), kv (b.kv), head (b.head), tail (b.tail), id (b.id), did (b.did), kakko (b.kakko), est_did (b.est_did), touten (b.touten), kuten (b.kuten), mlen (b.mlen), capacity (b.capacity) {};
    ~Bunsetsu () {}
    void init (char* cs) {
      cs = ny::skipChar (cs, ' ');
      char* p  = ny::skipChar (cs, ' ');
      id = ny::strton <int> (cs);
      cs = p;
      ++p; while (std::isdigit (*p) != 0) ++p;
      // dtype = *p;
      *p = '\0';
      did = ny::strton <int> (cs);
    }
    void addMorph (char* cs, const ssize_t &len) {
      if (mlen == capacity) {
        capacity *= 2;
        morph.resize (capacity, Morph (_dinfo, _verbosity, _output));
      }
      morph[mlen].setInfo (cs, len); ++mlen;
    }
    bool setInfo () { // set head & tail
      for (int i = 0; i < mlen; ++i) {
        const Morph &m = morph[i];
        if (m.cpos == _dinfo->sp_id) {
          if (m.fpos == _dinfo->touten_id) {
            touten |= true;
          } else if (m.fpos == _dinfo->kuten_id) {
            kuten  |= true;
          } else {
#ifdef USE_EMNLP_FEAT
            const Jmap::const_iterator it = _jk2id->find (m.surf);
            if (it != _jk2id->end ()) kv.push_back (it->second);
#endif
            if      (m.fpos == _dinfo->kakkoe_id) --kakko;
            else if (m.fpos == _dinfo->kakkos_id) ++kakko;
          }
        } else {
          tail = i;
#ifdef USE_EMNLP_FEAT
          if (m.cpos != _dinfo->jo_id && m.cpos != _dinfo->suf_id)
            head = i;
#else
          if (m.cpos != _dinfo->jo_id) {
            head = i; // +0.2
          } else {
            const Jmap::const_iterator it = _jk2id->find (m.surf);
            if (it != _jk2id->end ()) kv.push_back (it->second);
          }
#endif
        }
      }
#ifdef USE_EMNLP_FEAT
      if (tail != -1) {
        const Morph& m = morph[tail];
        if (m.cpos == _dinfo->jo_id) {
          const Jmap::const_iterator it = _jk2id->find (m.surf);
          if (it != _jk2id->end ()) kv.push_back (it->second);
        }
      }
#endif
      if (did < 0 || did > id) {
        return true;
      } else { // illegal bunsetsu annotation
        if (_verbosity > 0) {
          std::cerr << "\todd dependency.. " << id << "->" << did<< " ";
          if (_verbosity > 1) print ();
          std::cerr << std::endl;
        }
        return false;
      }
    }
    void print () const
    { for (int i = 0; i < mlen; ++i) std::cerr << morph[i].surf_str; }
  };

  class Sentence : private ny::Uncopyable {
  private:
    const DicInfo* _dinfo;
    const Jmap*    _jk2id;
    int            _verbosity;
    bool           _output;
    char           res[8192*640];
    char           *p;
  public:
    std::vector <Bunsetsu> bun;
    int                    len;
    int                    capacity;
    Sentence (ny::DA** da, const Jmap* jk2id, const ny::uint verbosity, const bool output = false)  : _dinfo (new DicInfo (da)), _jk2id (jk2id), _verbosity (verbosity), _output (output), p (&res[0]), bun (), len (0), capacity (1) {
      bun.resize (capacity, Bunsetsu (_dinfo, _jk2id, _verbosity, _output));
    };
    ~Sentence () { if (_dinfo) delete _dinfo; }
    void clear () {
      for (int i = 0; i < len; ++i) bun[i].clear ();
      len = 0;
    };
    void setHeader (const char* cs, const ssize_t &hlen) {
      std::memcpy (p, cs, hlen); p += hlen;
      std::memcpy (p, "\n", 1);  p += 1;
    }
    void setInfo () { // set info for all bunsetsu
      for (int i = 0; i < len; ++i) bun[i].setInfo (); // reset mlen, kv
    }
    void addBunsetsu (char* cs) {
      if (len == capacity)
        { capacity *= 2; bun.resize (capacity, Bunsetsu (_dinfo, _jk2id, _verbosity, _output)); }
      bun[len].init (cs); ++len;
    }
    void printParsed () {
      for (int i = 0; i < len; ++i) {
        const Bunsetsu &b = bun[i];
        p += sprintf (p, "* %d %dD\n", i, b.est_did);
        for (int j = 0; j < b.mlen; ++j) {
          const Morph &m = b.morph[j];
          std::memcpy (p, m.info_str, m.info_len); p += m.info_len;
          std::memcpy (p, "\n", 1); p += 1;
        }
      }
      std::memcpy (p, "EOS\n", 4); p += 4;
      char *check = &res[0];
      while (check < p) check += write (1, check, p - check);
      p = &res[0];
    }
    void print () { for (int i = 0; i < len; ++i) bun[i].print (); }
  };

  class DepParser : private ny::Uncopyable {
  private:
    enum learner_type    {ME, SVM, OLL};
    enum reg_type        {L1, L2}; // ME reguralization
    enum parser_type     {linear, chunking, backward, tournament};
#ifdef NDEBUG
    enum classifier_type {pME, pSVM, PA};
#else
    enum classifier_type {libME, pME, TinySVM, pSVM, PA};
#endif
    ny::Options _opt;
    // option variables for learners and classifiers
    int             _verbosity;       // verbosity level
    learner_type    _learner;        // learner
    classifier_type _classifier;     // classifier
    parser_type     _parser;         // parser
    process_t       _mode;           // mode
    std::string     _classifier_opt; // classifier options
    ny::uint        _cthr;           // cutoff threshold for unknown words
    ny::fl_t        _rcost;          // cost for regularization
#if defined (USE_SVM) || defined (USE_PA)
    pecco::PolyK*   _svm;
#endif
#ifdef USE_SVM
    TinySVM::Model* _tinysvm;
#endif
#ifdef USE_ME
    pecco::LLM*     _me;
    ME_Model        _libme;
    reg_type        _me_t;           // ME reguralizationt type (L1 or L2)
#endif
#ifdef USE_PA
    opal::Model*    _pa;
    int             _paiter;         // # iteration of PA
    opal::mem_pool  _ex_pool;
#endif
    ny::uint        _d;              // order of feature combination
    ny::uint        _lim;            // limit # of sent. processed
    ny::uint        _lim_cache;      // limit # of sent. processed (cache)
    // file names
    std::string     _dfn;            // dictionary file
    std::string     _jfn;            // dictionary file (joshi/kakko etc.)
    std::string     _mbase;          // ml data suffix
    std::string     _lfn;            // ml training data
    std::string     _cfn;            // ml cache data for fstrie
    std::string     _mfn;            // ml model data
    // result statistics
    ny::uint       _bnum;            // # total bunsetsu
    ny::uint       _bcorr;           // # correctly analyzed by parser
    ny::uint       _bbase;           // # correctly analyzed (baseline)
    ny::uint       _snum;            // # total sentence
    ny::uint       _scorr;           // # correctly analyzed by parser
    ny::uint       _sbase;           // # correctly analyzed (baseline)
    ny::uint       _cnum;            // total trials of classification
    // morph dictionary
    ny::uint       _word_id;         // index for values
    dic_t          _w2wi[NUM_FIELD]; // word * type -> word info
    char*          _ddap;            // memory keeper for double array
    ny::DA*        _dict_da[NUM_FIELD];
    std::vector <DictKey*>* _dakey_ary[NUM_FIELD];
    Jmap           _jk2id;
    // temporary variables to store events
    ny::fv_t       _fv;              // feature vector of the example
    Sentence*      _s;
    ny::uint       _global_offset;   // offset of global feature numbers
    charp_set      _jk_set;
    ny::uint       _jklen;           // # of joshi / # bit
    FILE*          _writer;
    std::stack <ny::uint> _stack;
    struct chunk_info {
      int  id;
      chunk_info (int id_) : id (id_) {};
      // bool done; // deterministic
      // chunk_info (int id_, bool done_) : id (id_), done (done_) {};
    };
    std::list <chunk_info> _cinfo;
    // timer
#ifdef USE_TIMER
    ny::TimerPool  _timer_pool;
    ny::Timer*     _io_t;
    ny::Timer*     _dict_t;
    ny::Timer*     _preproc_t;
    ny::Timer*     _parser_t;
    ny::Timer*     _classify_t;
#endif
    std::string _convertFv2Estr (const ny::fv_t &fv) const {
      std::ostringstream ss;
      ny::fv_it it = fv.begin ();
      if (! fv.empty ()) ss << *it << ":1";
      for (++it; it != fv.end (); ++it) ss << " " << *it << ":1";
      return ss.str ();
    }
    void _writeFv (const bool flag) const {
      std::fprintf (_writer, "%c1", flag ? '+' : '-');
      for (ny::fv_it it = _fv.begin (); it != _fv.end (); ++it)
        std::fprintf (_writer, " %d:1", *it);
      std::fprintf (_writer, "\n");
    }
    // feature vector generators
    void _set_global_feature (int i, int j, const Bunsetsu &bi, const Bunsetsu &bj, ny::uint &fi);
    void _set_local_feature (const Bunsetsu &bi, const int mid, ny::uint &offset);
    void _set_lexical_feature (int from, int to, ny::uint &fi);
    bool _event_gen_bi (int i, int j);
    bool _event_gen_tri (int i, int j, int k);
    //
    bool _init ();
    bool _setMorphDic ();
    bool _registerMorph (char* cs, const ssize_t &len);
    bool _learn ();
    void _processSample (const bool flag);
    template <process_t MODE> bool _analyze ();
    template <process_t MODE> bool _process ();
    template <process_t MODE> bool _processLinear ();
    template <process_t MODE> bool _processChunking ();
    template <process_t MODE> bool _processBackward ();
    // antisymmetric training / learning
    bool _eventOutputTournament ();
    bool _parseTournament ();
    bool _setOpt (int argc, char* argv[]) {
      std::vector <std::string> argvec (&argv[0], &argv[argc]);
      _opt.parse(argvec);
      // process mode
      if      (_opt["-m"] == "learn") _mode = learn;
      else if (_opt["-m"] == "cache") _mode = cache;
      else if (_opt["-m"] == "parse") _mode = parse;
      else if (_opt["-m"] == "both")  _mode = both;
      else ny::print_err (HERE "unknown mode: %s\n", _opt["-m"].c_str ());
      // classifier type
      if      (_opt["-c"] == "pSVM") _classifier = pSVM;
      else if (_opt["-c"] == "pME")  _classifier = pME;
      else if (_opt["-c"] == "PA")   _classifier = PA;
#ifndef NDEBUG
      else if (_opt["-c"] == "TinySVM") _classifier = TinySVM;
      else if (_opt["-c"] == "libME")   _classifier = libME;
#endif
      else ny::print_err (HERE "unknown classifier: %s\n", _opt["-c"].c_str ());
      // parsing algorithm
      if      (_opt["-p"] == "linear")     _parser = linear;
      else if (_opt["-p"] == "chunking")   _parser = chunking;
      else if (_opt["-p"] == "backward")   _parser = backward;
      else if (_opt["-p"] == "tournament") _parser = tournament;
      else ny::print_err (HERE "unknown parser: %s\n", _opt["-p"].c_str ());
      // option variables for learners and classifiers 
      _verbosity = ny::strton <int>      (_opt["-v"].c_str ());
      _cthr      = ny::strton <ny::uint> (_opt["-e"].c_str ());
      _rcost     = ny::strton <ny::fl_t> (_opt["-rc"].c_str ());
#ifdef USE_ME
      _me_t      = (_opt["-mt"] == "L1" ? L1 : L2);
#endif
#ifdef USE_PA
      _paiter    = ny::strton <ny::uint> (_opt["-oi"].c_str ());
#endif
      _d         = ny::strton <ny::uint> (_opt["-d"].c_str ());
      _lim       = ny::strton <ny::uint> (_opt["-n"].c_str ());
      _lim_cache = ny::strton <ny::uint> (_opt["-nn"].c_str ());
      // file names
      size_t pos = _opt["-lf"].rfind ("/");
      pos = (pos != std::string::npos) ? pos + 1 : 0;
      // check -model
      if (! opendir (_opt["-model"].c_str ())) {
        std::fprintf (stderr, "no such directory: %s [-model]\n", _opt["-model"].c_str ());
        std::exit (1);
      }
      _mbase = _opt["-model"] + "/" + _opt["-lf"].substr (pos) + "." + _opt["-n"];
      _dfn += _mbase + ".dic";
      _jfn += _mbase + ".joshi";
      // learner type
      std::string mid (""); // model id
      switch (_classifier) {
#ifndef NDEBUG
        case TinySVM:
#endif
        case pSVM: _learner = SVM; mid += ".SVM"; break;
#ifndef NDEBUG
        case libME:
#endif
        case pME:  _learner = ME;  mid += ".ME" + _opt["-mt"]; break;
        case PA:   _learner = OLL; mid += ".PA1";  break;
      }
      mid += ".e" + _opt["-e"] + ".d" + _opt["-d"] + ".c" + _opt["-rc"];
#ifdef USE_PA
      if (_learner == OLL) mid += ".i" + _opt["-oi"];
#endif
      mid += ".p" + _opt["-p"].substr (0, 1);
      _lfn += _mbase + mid + ".train";
      _mfn += _mbase + mid + ".model";
      _classifier_opt = "%0 " + _opt["-co"] + " -m " + _mfn + " -lf " + _lfn + " -v " + _opt["-v"];

      // cache file
      if (_opt["-cf"] != "") {
        pos = _opt["-cf"].rfind ("/");
        pos = (pos != std::string::npos) ? pos + 1 : 0;
        _cfn += _opt["-model"] + "/" + _opt["-cf"].substr (pos);
        _cfn += "." + _opt["-nn"] + "." + _opt["-n"] + mid + ".event";
      }
      if (_verbosity > 0)
        std::cerr << "classifier options: " << _classifier_opt << std::endl;
      if (_verbosity > 0) _opt.print ();

      return true;
    };
    // helper functions
    static void _printStatistics (const char* l, ny::uint bc, ny::uint bn, ny::uint sc, ny::uint sn) {
      fprintf (stderr, "%s % 3.4f (%5d/%5d) / % 3.4f (%5d/%5d)\n",
               l, bc * 100.0 / bn, bc, bn, sc * 100.0 / sn, sc, sn);
    }
#ifdef USE_ME
    static void _setExpanded (ny::fv_t& fv, const ny::uint d, ME_Sample &ms) {
#ifndef NDEBUG
      std::sort (fv.begin (), fv.end ()); // for libME
#endif
      char key[128];
      size_t fv_size = fv.size ();
      for (ny::uint fi = 0; fi < fv_size; ++fi) {
        int pos_i = 0;
        pos_i += sprintf (key, "%d", fv[fi]);
        ms.add_feature (key);
        if (d == 1) continue;
        for (ny::uint fj = fi + 1; fj < fv_size; ++fj) {
          int pos_j = pos_i;
          pos_j += sprintf (key + pos_i, ":%d", fv[fj]);
          ms.add_feature (key);
          if (d == 2) continue;
          for (ny::uint fk = fj + 1; fk < fv_size; ++fk) {
            sprintf (key + pos_j, ":%d", fv[fk]);
            ms.add_feature (key);
          }
        }
      }
    }
#endif
    ny::fl_t _classify (ny::fv_t &fv) {
#ifdef USE_ME
      if (_learner == ME) return _getMargin (fv);
#endif
#if defined (USE_SVM) || defined (USE_PA)
      return pecco::PolyK::sigmoid (_getMargin (fv));
#endif
      return 0;
    }
    bool _binClassify (ny::fv_t &fv) {
#ifdef USE_ME
      if (_learner == ME) return _getMargin (fv) > 0.5;
#endif
#if defined (USE_SVM) || defined (USE_PA)
      return _getMargin (fv) > 0;
#endif
      return false;
    }
    inline ny::fl_t _getMargin (ny::fv_t &fv) {
      ny::fl_t margin = 0;
      switch (_classifier) {
#if defined (USE_SVM) || defined (USE_PA)
        case pSVM: margin = _svm->classify (_fv); break;
#endif
#ifdef USE_ME
        case pME:  margin = _me->classify  (_fv); break;
#endif
#ifdef USE_PA
        case PA:   margin = _pa->classify (_fv); break;
#endif
#ifndef NDEBUG
#ifdef USE_SVM
        case TinySVM:
          margin = _tinysvm->classify (_convertFv2Estr (_fv).c_str ()); break;
#endif
#ifdef USE_ME
        case libME: {
          ME_Sample ms;
          _setExpanded (_fv, _d, ms);
          margin =  _libme.classify (ms)[_libme.get_class_id("+1")]; break;
        }
#endif
#endif
        default: assert (! "classifier disabled");
      }
      return margin;
    }
  public:
    DepParser (int argc, char* argv[]) :
      _opt ("J.DepP - Japanese Dependency Parser\nCopyright (c) 2008-2011 Naoki Yoshinaga\n"), _verbosity (0), _learner (OLL), _classifier (PA), _parser (linear), _mode (both), _classifier_opt ("-c pke -s 0.02"), _cthr (0), _rcost (static_cast <ny::fl_t> (0.00005)),
#if defined (USE_SVM) || defined (USE_PA)
      _svm (0),
#ifdef USE_SVM
      _tinysvm (0),
#endif
#endif
#ifdef USE_ME
      _me (0), _libme (), _me_t (L1),
#endif
#ifdef USE_PA
      _pa (0), _paiter (40), _ex_pool (),
#endif
      _d (2), _lim (0), _lim_cache (0), _dfn (""), _jfn (""), _mbase (""),  _lfn (""), _cfn (""), _mfn (""), _bnum (0), _bcorr (0), _bbase (0), _snum (0), _scorr (0), _sbase (0), _cnum (0), _word_id (FI_OFFSET), _ddap (0), _jk2id (), _fv (), _s (0), _global_offset (0), _jk_set (), _jklen (0), _writer (0), _stack (), _cinfo ()
#ifdef USE_TIMER
      , _timer_pool (), _io_t (_timer_pool.push ("io")), _dict_t (_timer_pool.push ("dict")), _preproc_t (_timer_pool.push ("preproc", "sent.")), _parser_t (_timer_pool.push ("parse", "sent.")), _classify_t (_timer_pool.push ("classify"))
#endif
    {
      // morph dict memory allocation
      for (ny::uint i = 0; i < NUM_FIELD; ++i) {
        _dict_da[i]   = new ny::DA;
        _dakey_ary[i] = new std::vector <DictKey*>;
      }
      // set default options
      _opt.set ("-lf", "train.KNP", "annotated sentences for training");
      _opt.set ("-tf", "-", "annotated sentences for testing; set '-' to read stdin");
      _opt.set ("-cf", "train.KNP", "preprocessed sentences to build fstrie");// fstrie
      _opt.set ("-n",  "0", "# of annotated sentences used to train (0: all)");
      _opt.set ("-nn", "0", "# of preprocessed sentences used to build fstrie (0: all)");
      std::string dc_s;
      std::string c_s;
      std::string rc_s;
#ifdef USE_SVM
      dc_s  = "pSVM";
      c_s  += (c_s.empty  () ? "" : " / "); c_s  += "pSVM";
      rc_s += (rc_s.empty () ? "" :  " / "); rc_s += "SVM C";
#ifdef NDEBUG
      c_s  += " / TinySVM";
#endif
#endif
#ifdef USE_ME
      if (dc_s.empty ()) dc_s = "pME";
      c_s  += (c_s.empty  () ? "" : " / "); c_s  += "pME";
      rc_s += (rc_s.empty () ? "" : " / "); rc_s  += "ME L1 omega / ME L2 C";
#ifdef NDEBUG
      c_s  += " / libME";
#endif
#endif
#ifdef USE_PA
      if (dc_s.empty ()) dc_s = "PA"; 
      c_s  += (c_s.empty  () ? "" : " / "); c_s  += "PA";
      rc_s += (rc_s.empty () ? "" : " / "); rc_s += "PA C";
#endif
      _opt.set ("-c",     dc_s, "classifier: " + c_s);
      _opt.set ("-d",     "2", "degree of feature conjunction");
      _opt.set ("-rc",    "0.00005", rc_s);
#ifdef USE_PA
      _opt.set ("-oi",    "40", "PA iteration");
#endif
#ifdef USE_SVM
      _opt.set ("-s",     "1", "s in polynomial kernel (SVM)");
      _opt.set ("-r",     "1", "r in polynomial kernel (SVM)");
#endif
#ifdef USE_ME
      _opt.set ("-mt",    "L1", "ME regularization L1 / L2");
#endif
      _opt.set ("-co",    "-c pke -s 0.02", "classifier options");
      _opt.set ("-p",     "linear", "parsing algorithm; linear / chunking / backward / tournament");
      _opt.set ("-m",     "parse", "parser mode; learn / parse / both / cache");
      _opt.set ("-e",     "0", "morphological feature threshold");
      _opt.set ("-model", "model", "directory to store a model and temporary data");
      _opt.set ("-v",     "0", "verbosity; set to -1 to see output");
      // set options
      _setOpt (argc, argv);
      _init ();
    };
    ~DepParser () {
      for (ny::uint i = 0; i < NUM_FIELD; ++i) {
        for (dic_t::iterator it = _w2wi[i].begin ();
             it != _w2wi[i].end (); ++it)
          if (it->first != UNKNOWN) delete [] it->first;
        delete _dict_da[i];
        for (std::vector <DictKey*>::iterator it = (*_dakey_ary[i]).begin ();
             it != (*_dakey_ary[i]).end (); ++it)
          delete *it;
        delete _dakey_ary[i];
      }
      delete [] _ddap;
    };
    // interface
    bool run ();
    void parseInit () {
      switch (_classifier) {
#ifdef USE_SVM
        case pSVM:
          _svm = new pecco::PolyK (_classifier_opt.c_str ());
          _cfn += ".s" + _svm->getSigma ();
          _svm->setOpt ("-cf", _cfn.c_str ());
          _svm->load ();
          break;
#endif
#ifdef USE_ME
        case pME:
          _me = new pecco::LLM (_classifier_opt.c_str ());
          _me->setOpt ("-cf", _cfn.c_str ());
          _me->load ();
          break;
#endif
#ifdef USE_PA
        case PA:
          if (_d > 0) {
            // use polyK for Kernel
            _svm = new pecco::PolyK (_classifier_opt.c_str ());
            _cfn += ".s" + _svm->getSigma ();
            _svm->setOpt ("-cf", _cfn.c_str ());
            _svm->load ();
            _classifier = pSVM; // switch
          } else {
            _pa = new opal::Model (_d);
            _pa->load (_mfn.c_str ());
          }
          break;
#endif
#ifndef NDEBUG
#ifdef USE_SVM
        case TinySVM:
          _tinysvm = new TinySVM::Model;
          _tinysvm->read (_mfn.c_str ());
          break;
#endif
#ifdef USE_ME
        case libME: _libme.load_from_file (_mfn); break;
#endif
#endif
        default: assert (! "classifier disabled");
      }
    }
    void process ();
    Sentence* setSentence (const bool output)
    { return _s =  new Sentence (&_dict_da[0], &_jk2id, _verbosity, output); }
  };
}
#endif /* PDEP_H */
